"""This module contains the function to create a configuration file."""
import configparser
from pathlib import Path
from typing import Optional

from sprynger.utils.constants import CONFIG_FILE


def create_config(config_dir: Optional[Path] = None,
                  keys: Optional[list[str]] = None
                  ):
    """Initiates process to generate configuration file.

    :param keys: If you provide a list of keys, sprynger will skip the prompt.
    """
    from sprynger.utils.constants import DEFAULT_PATHS

    if not config_dir:
        config_dir = CONFIG_FILE

    config = configparser.ConfigParser()
    config.optionxform = str
    print(f"Creating config file at {config_dir} with default paths...")

    # Set directories
    config.add_section('Directories')
    for api, path in DEFAULT_PATHS.items():
        config.set('Directories', api, str(path))

    # Set authentication
    config.add_section('Authentication')
    if keys:
        if not isinstance(keys, list):
            raise ValueError("Parameter `keys` must be a list.")
        key = ", ".join(keys)
    else:
        prompt_key = "Please enter your API Key(s), obtained from "\
                     "https://dev.springernature.com.  Separate "\
                     "multiple keys by comma:\n"
        key = input(prompt_key)
    config.set('Authentication', 'APIKey', key)

    # Set default values
    config.add_section('Requests')
    config.set('Requests', 'Timeout', '20')
    config.set('Requests', 'Retries', '5')
    config.set('Requests', 'BackoffFactor', '2.0')

    # Write out
    config_dir.parent.mkdir(parents=True, exist_ok=True)
    with open(config_dir, "w") as ouf:
        config.write(ouf)
    print(f"Configuration file successfully created at {config_dir}\n"
          "For details see https://sprynger.rtfd.io/en/stable/initialization.html.")
    return config



"""Tests for the Meta class."""
from sprynger import init
from sprynger import Meta
from sprynger.utils.data_structures import (
    MetadataCreator,
    MetaDiscipline,
    MetadataFacets,
    MetaRecord,
    MetaURL,
)

init()

article = Meta(doi='10.1007/s00394-024-03496-7', refresh=True)

def test_results():
    """Test the results."""
    assert article.results.total == 1
    assert article.results.start == 1
    assert article.results.pageLength == 10
    assert article.results.recordsRetrieved == 1

def test_records():
    """Test the records."""
    expected_record = MetaRecord(
        contentType='Article',
        identifier='doi:10.1007/s00394-024-03496-7',
        language='en',
        urls=[
            MetaURL(
                format='html',
                platform='web',
                value='http://link.springer.com/openurl/fulltext?id=doi:10.1007/s00394-024-03496-7',
            ),
            MetaURL(
                format='pdf',
                platform='web',
                value='http://link.springer.com/openurl/pdf?id=doi:10.1007/s00394-024-03496-7',
            ),
            MetaURL(
                format='',
                platform='',
                value='http://dx.doi.org/10.1007/s00394-024-03496-7',
            ),
        ],
        title='Ultra-processed food intake in toddlerhood and mid-childhood in the UK: cross sectional and longitudinal perspectives',
        creators=[
            MetadataCreator(creator='Conway, Rana E.', ORCID='0000-0003-0955-7107'),
            MetadataCreator(creator='Heuchan, Gabriella N.', ORCID=None),
            MetadataCreator(creator='Heggie, Lisa', ORCID='0000-0002-4846-2357'),
            MetadataCreator(creator='Rauber, Fernanda', ORCID='0000-0001-9693-7954'),
            MetadataCreator(creator='Lowry, Natalie', ORCID='0000-0002-9137-5005'),
            MetadataCreator(creator='Hallen, Hannah', ORCID=None),
            MetadataCreator(creator='Llewellyn, Clare H.', ORCID='0000-0002-0066-2827'),
        ],
        publicationName='European Journal of Nutrition',
        openaccess=True,
        doi='10.1007/s00394-024-03496-7',
        publisher='Springer',
        publicationDate='2024-10-04',
        publicationType='Journal',
        issn='1436-6207',
        eIssn='1436-6215',
        volume='',
        number='',
        issueType='',
        topicalCollection='',
        genre=['OriginalPaper', 'Original Contribution'],
        startingPage=1,
        endingPage=12,
        journalId=394,
        onlineDate='2024-10-04',
        copyright='©2024 The Author(s)',
        abstract='Purpose (i) Characterize ultra-processed food (UPF) intakes in toddlerhood and mid-childhood, including identifying principal UPF sub-groups and associations with nutrient profile; (ii) explore stability and change in UPF intake between toddlerhood and mid-childhood. Methods Data were from children in the UK Gemini twin cohort at 21 months ( n \u2009=\u20092,591) and 7 years ( n \u2009=\u2009592) of age. UPF intakes were estimated using diet diaries and Nova classification. Complex samples general linear or logistic regression models were used to explore associations between UPF intake, UPF sub-groups and nutrients, and changes in intake over time. Results The contribution of UPF to total energy was 46.9% (±\u200914.7) at 21 months and 59.4% (±\u200912.5) at 7 years. Principal UPF sub-groups were yogurts, higher-fiber breakfast cereals, and wholegrain breads in toddlerhood, and puddings and sweet cereal products and white breads in mid-childhood. At both ages, mean free sugar and sodium intakes exceeded recommended maximums and higher UPF consumption was associated with consuming more of each nutrient ( P \u2009<\u20090.001). UPF intake was negatively associated with fat, saturated fat and protein intake in toddlerhood, and fiber intake in mid-childhood ( P \u2009<\u20090.001). Being in the highest UPF intake quintile in toddlerhood was predictive of being in the highest quintile in mid-childhood (OR 9.40, 95%CI 3.94–22.46). Conclusions UPF accounted for nearly half of toddlers’ energy, increasing to 59% in mid-childhood. Higher UPF consumers had higher intakes of free sugar and sodium. UPF intake in toddlerhood was predictive of mid-childhood intake. Effective policies are needed to reduce UPF intakes in the early years.',
        conferenceInfo=[],
        keyword=['Ultra-processed foods', 'Diet quality', 'Toddlers', 'Children', 'UK'],
        subjects=['Chemistry', 'Nutrition'],
        disciplines=[MetaDiscipline(id='3524', term='Nutrition')],
    )
    assert article.records[0] == expected_record

def test_facets():
    """Test the facets."""
    assert len(article.facets) == 12

    expected_first_facet = MetadataFacets(facet='subject', value='Chemistry', count=1)
    assert article.facets[0] == expected_first_facet



"""Module with auxiliary functions to parse OpenAccess documents."""
from lxml.etree import _Element

from sprynger.utils.data_structures import Affiliation, Contributor, Date, Paragraph
from sprynger.utils.parse import get_attr, get_text, make_int_if_possible


def affs_to_dict(affs) -> dict[str, Affiliation]:
    """Auxiliary function to query the affiliations by their number."""
    return {aff.nr: aff for aff in affs}


def get_affiliations(data: _Element) -> list[Affiliation]:
    """Parse the affiliations of the document."""
    affiliations = []
    for contrib_group in data.findall('.//contrib-group'):
        contribution_group = contrib_group.get('content-type')
        for a in contrib_group.findall('.//aff'):
            institution = a.find('.//institution-wrap')
            new_aff = Affiliation(
                type=contribution_group,
                ref_nr=a.get('id'),
                ror=get_attr(institution, 'institution-id', 'institution-id-type', 'ROR'),
                grid=get_attr(institution, 'institution-id', 'institution-id-type', 'GRID'),
                isni=get_attr(institution, 'institution-id', 'institution-id-type', 'ISNI'),
                division=get_attr(institution, 'institution', 'content-type', 'org-division'),
                name=get_attr(institution, 'institution', 'content-type', 'org-name'),
                city=get_attr(a, 'addr-line', 'content-type', 'city'),
                country=get_text(a, './/country')
            )
            affiliations.append(new_aff)
    return affiliations


def get_contributors(data: _Element) -> list[Contributor]:
    """Parse the contributors of the document and matcg them with their affiliations."""
    contributors = []
    for c in data.findall('.//contrib'):
        # Get affiliation
        affs_nr = []
        for aff_ref in c.findall('.//xref[@ref-type="aff"]'):
            aff_nr = aff_ref.get('rid')
            affs_nr.append(aff_nr)

        # Get contributor data
        new_contrib = Contributor(
            type=c.get('contrib-type'),
            nr=c.get('id'),
            orcid=get_attr(c, 'contrib-id', 'contrib-id-type', 'orcid'),
            surname=get_text(c, './/name/surname'),
            given_name=get_text(c, './/name/given-names'),
            email=get_text(c, './/email'),
            affiliations_ref_nr=affs_nr
        )
        contributors.append(new_contrib)
    return contributors


def get_date(date_node: _Element) -> Date:
    """Auxiliary function to extract date information from a date node."""
    return Date(day=make_int_if_possible(get_text(date_node, './/day')),
                month=make_int_if_possible(get_text(date_node, './/month')),
                year=make_int_if_possible(get_text(date_node, './/year')))


def get_paragraphs(xml) -> list[Paragraph]:
    """Paragraphs of the OpenAccess document.

    Returns:
        list[Paragraph]: A list of Paragraph objects containing the 
        `paragraph_id`, `section_id`, `section_title`, and `text`.
    """
    if xml is not None:
        parsed_paragraphs = []
        for paragraph in xml.findall('.//p[@id]'):
            paragraph_id = paragraph.get('id')
            paragraph_text = ''.join(paragraph.itertext())

            parent = paragraph.getparent()
            section_id = parent.get('id')
            section_title = get_text(parent, 'title')


            paragraph_data = Paragraph(
                paragraph_id=paragraph_id,
                section_id=section_id,
                section_title=section_title,
                text=paragraph_text.strip()
            )
            parsed_paragraphs.append(paragraph_data)

        return parsed_paragraphs
    return []



"""Module with the chapter class for the OpenAccess class."""
from typing import Optional, Union

from sprynger.utils.data_structures import Affiliation, Contributor, Date, Paragraph
from sprynger.utils.parse import get_attr, get_text, make_int_if_possible
from sprynger.utils.parse_openaccess import (
    affs_to_dict,
    get_contributors,
    get_affiliations,
    get_date,
    get_paragraphs,
)

class Chapter:
    """Auxiliary class to parse a chapter from a book."""
    @property
    def affiliations(self) -> list[Affiliation]:
        """List of affiliations of the collaborators of the chapter. Each affiliation is represented
        as a named tuple with the following fields:
        `type`, `ref_nr`, `ror`, `grid`, `isni`, `division`, `name`, `city`, `country`.
        
        Note: To match affiliations with contributors use the affiliation's `ref_nr` and the
        contributor's `affiliations_ref_nr`.
        """
        return get_affiliations(self._data)

    @property
    def affiliations_dict(self) -> dict[str, Affiliation]:
        """Auxiliary property to query the affiliations by their reference number."""
        return affs_to_dict(self.affiliations)

    @property
    def contributors(self) -> list[Contributor]:
        """List of contributors of the chapter. Each contributor is represented as a named tuple
        with the following fields:
        `type`, `nr`, `orcid`, `surname`, `given_name`, `email`, `affiliations_ref_nr`.

        Note: To match contributors with affiliations use the contributor's `affiliations_ref_nr`
        and the affiliation's `ref_nr`.
        """
        return get_contributors(self._data)

    @property
    def book_doi(self) -> Optional[str]:
        """DOI of the book."""
        return get_attr(self._book_meta, 'book-id', 'book-id-type', 'doi')

    @property
    def book_pub_date(self) -> Optional[str]:
        """Publication date of the book."""
        return get_text(self._book_meta, './/pub-date[@date-type="pub"]/string-date')

    @property
    def book_title(self) -> Optional[str]:
        """Title of the book."""
        return get_text(self._book_meta, './/book-title-group/book-title')

    @property
    def book_title_id(self) -> Optional[str]:
        """Book title ID."""
        return get_attr(self._book_meta, 'book-id', 'book-id-type', 'book-title-id')

    @property
    def book_sub_title(self) -> Optional[str]:
        """Sub-title of the book."""
        return get_text(self._book_meta, './/book-title-group/subtitle')

    @property
    def chapter_nr(self) -> Optional[Union[int, str]]:
        """Book chapter name or number."""
        chapter_nr = get_attr(self._chapter_meta, 'book-part-id', 'book-part-id-type', 'chapter')
        return make_int_if_possible(chapter_nr)

    @property
    def date_epub(self) -> Date:
        """Electronic publication date of the chapter."""
        date_node = self._chapter_meta.find('.//pub-date[@publication-format="electronic"]')
        return get_date(date_node)

    @property
    def date_ppub(self) -> Date:
        """Print publication date of the chapter."""
        date_node = self._chapter_meta.find('.//pub-date[@publication-format="print"]')
        return get_date(date_node)

    @property
    def date_registration(self) -> Date:
        """Registration date of the chapter."""
        date_node = self._chapter_meta.find('.//pub-history/date[@date-type="registration"]')
        return get_date(date_node)

    @property
    def date_online(self) -> Date:
        """Online date of the chapter."""
        date_node = self._chapter_meta.find('.//pub-history/date[@date-type="online"]')
        return get_date(date_node)

    @property
    def doi(self) -> Optional[str]:
        """DOI of the chapter."""
        doi = get_attr(self._chapter_meta, 'book-part-id', 'book-part-id-type', 'doi')
        return doi

    @property
    def isbn_electronic(self) -> Optional[str]:
        """ISBN of the electronic version of the book."""
        return get_attr(self._book_meta, 'isbn', 'content-type', 'epub')

    @property
    def isbn_print(self) -> Optional[str]:
        """ISBN of the print version of the book."""
        return get_attr(self._book_meta, 'isbn', 'content-type', 'ppub')

    @property
    def paragraphs(self) -> list[Paragraph]:
        """Paragraphs of the book chapter.
        
        Returns:
            list[Paragraph]: A list of Paragraph objects containing the
            `paragraph_id`, `section_id`, `section_title`, and `text`.
        """
        return get_paragraphs(self._data)

    @property
    def publisher_id(self) -> Optional[str]:
        """Publisher ID of the chapter's book."""
        return get_attr(self._book_meta, 'book-id', 'book-id-type', 'publisher-id')

    @property
    def publisher_loc(self) -> Optional[str]:
        """Location of the publisher."""
        return get_text(self._book_meta, './/publisher/publisher-loc')

    @property
    def publisher_name(self) -> Optional[str]:
        """Name of the publisher."""
        return get_text(self._book_meta, './/publisher/publisher-name')

    @property
    def title(self) -> Optional[str]:
        """Title of the chapter."""
        return get_text(self._chapter_meta, './/title-group/title')

    def __init__(self, data):
        self._data = data
        self._book_meta = data.find('.//book-meta')
        self._chapter_meta = data.find('.//book-part[@book-part-type="chapter"]/book-part-meta')


    def __repr__(self) -> str:
        return f'Chapter {self.doi}'


"""Module with all the data structures used in the package."""
from collections import namedtuple

def create_namedtuple(name: str, fields: list, defaults=None):
    """Create a namedtuple with default values."""
    default_list = [defaults] * len(fields)
    return namedtuple(name, fields, defaults=default_list)

#############################
#          Metadata         #
#############################
fields_metadata_results = ['total', 'start', 'pageLength', 'recordsRetrieved']
MetadataResult = create_namedtuple('MetadataResult', fields_metadata_results)

fields_metadata_record = ['contentType', 'identifier', 'language', 'url',
                          'url_format', 'url_platform' ,'title', 'creators',
                          'publicationName', 'openaccess', 
                          'doi', 'publisher', 'publicationDate',
                          'publicationType', 'issn', 'volume', 'number', 'genre', 
                          'startingPage', 'endingPage', 'journalId',
                          'copyright', 'abstract', 'subjects']
MetadataRecord = create_namedtuple('MetadataRecord', fields_metadata_record)

fields_metadata_creator = ['creator', 'ORCID']
MetadataCreator = create_namedtuple('MetadataCreator', fields_metadata_creator)

fields_metadata_facets = ['facet', 'value', 'count']
MetadataFacets = create_namedtuple('MetadataFacets', fields_metadata_facets)

#############################
#         Open Access       #
#############################

# Open Access Paragraph
fields_oa_paragraphs = ['paragraph_id', 'section_id', 'section_title', 'text']
Paragraph = create_namedtuple('Paragraph', fields_oa_paragraphs)

fields_oa_contributor = ['type', 'nr', 'orcid', 'surname', 'given_name', 'email', 'affiliations_ref_nr']
Contributor = create_namedtuple('Contributor', fields_oa_contributor)

fields_oa_aff = ['type', 'ref_nr', 'ror', 'grid', 'isni', 'division', 'name', 'city', 'country']
Affiliation = create_namedtuple('Affiliation', fields_oa_aff)

fields_date = ['year', 'month', 'day']
Date = create_namedtuple('Date', fields_date)

#############################
#         Meta             #
#############################

fields_meta_url = ['format', 'platform', 'value']
MetaURL = create_namedtuple('MetaURL', fields_meta_url)

fields_meta_discipline = ['id', 'term']
MetaDiscipline = create_namedtuple('MetaDiscipline', fields_meta_discipline)

fields_meta_record = [
    'contentType', 'identifier', 'language',
    'urls', 'title', 'creators',
    'publicationName', 'openaccess', 'doi',
    'publisher', 'publicationDate', 'publicationType',
    'issn', 'eIssn', 'volume',
    'number', 'issueType', 'topicalCollection',
    'genre', 'startingPage', 'endingPage',
    'journalId', 'onlineDate', 'copyright',
    'abstract', 'conferenceInfo', 'keyword',
    'subjects', 'disciplines'
]
MetaRecord = create_namedtuple('MetaRecord', fields_meta_record)



"""Module to test the Retrieve class."""
import pytest

from sprynger import init
from sprynger.retrieve import Retrieve

init()

def test_check_query():
    """Test the _check_query method."""
    with pytest.raises(ValueError, match="Please provide a query or kwargs."):
        Retrieve('', 'OpenAccess')

    error_str = f"Invalid field: erroneus_field."
    with pytest.raises(ValueError, match=error_str):
        Retrieve('', api='Meta', erroneus_field='Neural Network')

    error_str = f"Field topicalcollection is not available in OpenAccess."
    with pytest.raises(ValueError, match=error_str):
        Retrieve('', api='OpenAccess', topicalcollection='Neural Network')

    error_str = f"Field topicalcollection is not available in Basic plan."
    with pytest.raises(ValueError, match=error_str):
        Retrieve('', api='Meta', premium=False, topicalcollection='Neural Network')



"""Module with the Retrieval Class"""
from typing import Optional, Literal, Union

from sprynger.base import Base
from sprynger.utils.constants import VALID_FIELDS


class Retrieve(Base):
    """Retrieve data from the Springer API."""
    def _check_query(self,
                    query: str,
                    kwargs: Optional[dict],
                    api: Literal['OpenAccess', 'Meta', 'Metadata'],
                    plan: Literal['Basic', 'Premium']) -> None:
        """Auxiliary function to check if the query and kwargs are valid."""
        # Either query or kwargs should be provided
        self._validate_presence_of_query_or_kwargs(query, kwargs)
        # Check if kwargs are valid
        if kwargs:
            self._validate_kwargs(kwargs, api, plan)

    def _validate_presence_of_query_or_kwargs(self, query: str, kwargs: Optional[dict]) -> None:
        """Ensure that either query or kwargs are provided."""
        if not query and not kwargs:
            raise ValueError("Please provide a query or kwargs.")

    def _validate_kwargs(self, kwargs: dict, api: str, plan: str) -> None:
        """Validate the provided kwargs against the valid fields."""
        for field in kwargs:
            self._validate_field(field, api, plan)

    def _validate_field(self, field: str, api: str, plan: str) -> None:
        """Validate a single field against the valid fields."""
        if field not in VALID_FIELDS:
            raise ValueError(f"Invalid field: {field}.")
        if api not in VALID_FIELDS[field].get('api', []):
            raise ValueError(f"Field {field} is not available in {api}.")
        if plan not in VALID_FIELDS[field].get('plan', []):
            raise ValueError(f"Field {field} is not available in {plan} plan.")


    def _build_filters(self, kwargs: dict) -> str:
        """Auxiliary function to build filters for the query."""
        filters = ''
        if kwargs:
            for field, value in kwargs.items():
                if isinstance(value, list):
                    filters += ' '.join(f'{field}:{v}' for v in value)
                else:
                    filters = self._and_join(filters, f'{field}:{value}')
        return filters

    def _and_join(self,
                  str1: str,
                  str2: str) -> str:
        """Auxiliary function to join two strings with AND"""
        if str1 and str2:  # Both strings are non-empty
            return f"{str1} {str2}"
        return str1 or str2


    def __init__(self,
                 query: str,
                 api: Literal['Metadata', 'Meta', 'OpenAccess'],
                 start: int = 1,
                 nr_results: int = 10,
                 premium: bool = False,
                 cache: bool = True,
                 refresh: Union[bool, int] = False,
                 **kwargs):
        """This class handles the query to retrieve the data from the Springer API."""

        plan = 'Premium' if premium else 'Basic'
        self._check_query(query, kwargs, api, plan)
        filters = self._build_filters(kwargs)
        query = self._and_join(query, filters)

        super().__init__(query=query,
                         api=api,
                         start=start,
                         nr_results=nr_results,
                         premium=premium,
                         cache=cache,
                         refresh=refresh)



"""
Module with Metadata class.
"""
from typing import Union

from sprynger.retrieve import Retrieve
from sprynger.utils.data_structures import (MetadataCreator,
                                            MetadataFacets,
                                            MetadataRecord,
                                            MetadataResult)
from sprynger.utils.parse import make_int_if_possible, str_to_bool


class Metadata(Retrieve):
    """Class to retreive the metadata of a document from the Springer Metadata API."""
    @property
    def facets(self) -> list[MetadataFacets]:
        """Faceted information about the results.

        Returns:
            list[MetadataFacets]: A list of MetadataFacets objects containing the 
            `facet`, `value`, and `count`.
        """
        facets_list = []
        factets = self._json.get('facets', [])
        for facet in factets:
            facet_name = facet.get('name')
            for item in facet.get('values', []):
                new_facet = MetadataFacets(facet=facet_name,
                                           value=item.get('value'),
                                           count=make_int_if_possible(item.get('count')))
                facets_list.append(new_facet)
        return facets_list

    @property
    def json(self) -> dict:
        """Raw JSON response from the Springer Metadata API."""
        return self._json

    @property
    def results(self) -> MetadataResult:
        """Overview of the results of the query.

        Returns:
            MetadataResult: An object containing the the `total` matches found, `start` 
            index of the first result, max `pageLength` when paginating and the number of 
            `recordsRetrieved`.
        """
        res = self._json.get('result', [{}])[0]
        total_results = int(res.get('total', 0))
        nr_results = min(self._nr_results, total_results)
        out = MetadataResult(total=total_results,
                             start=int(res.get('start', 0)),
                             pageLength=int(res.get('pageLength', 0)),
                             recordsRetrieved=nr_results)
        return out

    @property
    def records(self) -> list[MetadataRecord]:
        """Contains the individual records that matched the query.

        Returns:
            list[MetadataRecord]: List of MetadataRecord objects which contain the following
            items of a document: `contentType`, 
            `identifier`, `language`, `url`, `url_format`, `url_platform`, `title`, `creators`, 
            `publicationName`, `openaccess`, `doi`, `publisher`, `publicationDate`,
            `publicationType`, `issn`, `volume`, `number`, `genre`, `startingPage`, 
            `endingPage`, `journalId`, `copyright`, `abstract` and `subjects`.
        """
        records_list = []
        for record in self._json.get('records', []):
            url = record.get('url', {})[0].get('value')
            url_format = record.get('url', {})[0].get('format')
            url_platform = record.get('url', {})[0].get('platform')

            creators = []
            for ceator in record.get('creators', []):
                creators.append(MetadataCreator(creator=ceator.get('creator'),
                                                ORCID=ceator.get('ORCID')))

            records_list.append(
                MetadataRecord(
                    contentType=record.get('contentType'),
                    identifier=record.get('identifier'),
                    language=record.get('language'),
                    url=url,
                    url_format=url_format,
                    url_platform=url_platform,
                    title=record.get('title'),
                    creators=creators,
                    publicationName=record.get('publicationName'),
                    openaccess=str_to_bool(record.get('openaccess')),
                    doi=record.get('doi'),
                    publisher=record.get('publisher'),
                    publicationDate=record.get('publicationDate'),
                    publicationType=record.get('publicationType'),
                    issn=record.get('issn'),
                    volume=make_int_if_possible(record.get('volume')),
                    number=make_int_if_possible(record.get('number')),
                    genre=record.get('genre'),
                    startingPage=make_int_if_possible(record.get('startingPage')),
                    endingPage=make_int_if_possible(record.get('endingPage')),
                    journalId=make_int_if_possible(record.get('journalId')),
                    copyright=record.get('copyright'),
                    abstract=record.get('abstract'),
                    subjects=record.get('subjects')
                )
            )
        return records_list

    def __init__(self,
                 query: str = '',
                 start: int = 1,
                 nr_results: int = 10,
                 premium: bool = False,
                 cache: bool = True,
                 refresh: Union[bool, int] = False,
                 **kwargs):
        """
        Args:
            query (str): The query to search for.
            start (int): The starting index for the results. Defaults to 1.
            nr_results (int): The number of results to retrieve. Defaults to 10.
            premium (bool): Whether the user has a premium account. Defaults to False.
            cache (bool): Whether to cache the results. Defaults to True.
            refresh (bool|int): Weather to refresh the cache. If an integer is provided, 
                it will be used as the cache expiration time in days. Defaults to False.
            kwargs: Additional fields for query (e.g. issn, datefrom, dateto, etc.). For a comprehensive list of
                available fields, see the 
                `Springer Metadata API documentation <https://docs-dev.springernature.com/docs/#querying-api/querying-api>`_.

        This class is iterable, allowing you to iterate over the metadata `records` retrieved.
        It also supports indexing to access the metadata of specific documents.

        Example:
            Retrieve the metadata of documents with the word 'Segmentation' from the journal
            with the ISSN '1573-7497' starting from the date '2024-01-01'.
            

            >>> metadata = Metadata('Segmentation', issn='1573-7497', datefrom='2024-01-01')
            >>> for record in metadata:
            >>>     print(record)
     
        Note:
            The properties `facets`, `records` and `results` can be converted to a pandas 
            DataFrame with `pd.DataFrame(object.property)`.
        """
        api = self.__class__.__name__
        super().__init__(query=query,
                         api=api,
                         start=start,
                         nr_results=nr_results,
                         premium=premium,
                         cache=cache,
                         refresh=refresh,
                         **kwargs)
        self._nr_results = nr_results
        self._records = self.records

    def __iter__(self):
        return iter(self._records)

    def __getitem__(self, index):
        return self._records[index]

    def __len__(self):
        return len(self._records)

    def __repr__(self):
        return self._records.__repr__()



from configparser import ConfigParser, NoSectionError

from pathlib import Path
from typing import List, Optional, Union

from sprynger.utils.constants import CONFIG_FILE, DEFAULT_PATHS
from sprynger.utils.create_config import create_config

CONFIG = None
CUSTOM_KEYS = None

def init(config_dir: Union[str, Path] = CONFIG_FILE,
         keys: Optional[List[str]] = None) -> None:
    """
    Function to initialize the sprynger library. For more information go to the
    `documentation <file:///Users/nilsherrmann/sprynger/docs/build/html/initialization.html#configuration>`_.
    
    Args:
        config_dir (str): Path to the configuration file
        keys (list): List of API keys
    
    Raises:
        FileNotFoundError: If the configuration file is not found.
    
    Example:
        >>> from sprynger import init
        >>> init(config_dir='path/to/custom/config.cfg', keys=['key1', 'key2'])
    """
    global CONFIG
    global CUSTOM_KEYS

    config_dir = Path(config_dir)

    if not config_dir.exists():
        CONFIG = create_config(config_dir, keys)
    else:
        CONFIG = ConfigParser()
        CONFIG.optionxform = str
        CONFIG.read(config_dir)

    check_sections(CONFIG)
    check_default_paths(CONFIG, config_dir)
    create_cache_folders(CONFIG)

    CUSTOM_KEYS = keys


def check_sections(config: ConfigParser) -> None:
    """Auxiliary function to check if all sections exist."""
    for section in ['Directories', 'Authentication']:
        if not config.has_section(section):
            raise NoSectionError(section)


def check_default_paths(config: ConfigParser,
                        config_path: Path) -> None:
    """Auxiliary function to check if default cache paths exist.
    If not, the paths are writen in the config.
    """
    for api, path in DEFAULT_PATHS.items():
        if not config.has_option('Directories', api):
            config.set('Directories', api, str(path))
            with open(config_path, 'w', encoding='utf-8') as ouf:
                config.write(ouf)


def create_cache_folders(config: ConfigParser) -> None:
    """Auxiliary function to create cache folders."""
    for _, path in config.items('Directories'):
        cache_path = Path(path)
        cache_path.mkdir(parents=True, exist_ok=True)


def get_config() -> ConfigParser:
    """Function to get the config parser."""
    if not CONFIG:
        raise FileNotFoundError('No configuration file found.'
                                'Please initialize sprynger with init().\n'
                                'For more information visit: '
                                'the documentation.')
    return CONFIG


def get_keys() -> List[str]:
    """Function to get the API keys and overwrite keys in config if needed."""
    if CUSTOM_KEYS:
        keys = CUSTOM_KEYS
    else:
        keys = [k.strip() for k in CONFIG.get('Authentication', 'APIKey').split(",")]
    return keys





"""Tests for the OpenAccess class."""
from sprynger import init, OpenAccess
from sprynger.openaccess import Article, Chapter
from sprynger.utils.data_structures import Affiliation, Contributor, Date, Paragraph

init()

book = OpenAccess(isbn="978-3-031-63500-7", start=1, nr_results=2, refresh=30)
chapter = OpenAccess("doi:10.1007/978-3-031-61874-1_5", refresh=30)
chapter_with_text = OpenAccess(doi="10.1007/978-3-031-24498-8_7", refresh=30)
book_pagination = OpenAccess(isbn='978-3-031-63498-7', nr_results=30, refresh=30)

journal = OpenAccess(issn="2198-6053", start=4, nr_results=3, refresh=30)
journal_pagination = OpenAccess('issn:2198-6584', nr_results=26, refresh=True)
article = OpenAccess(doi="10.1007/s40747-024-01577-y", refresh=30)


def test_article_affiliations():
    """Test the affiliations of the article."""
    expected = [
        Affiliation(type=None, ref_nr='Aff1', ror='https://ror.org/04z7qrj66', grid='grid.412518.b', isni='0000 0001 0008 0619', division='College of Information Engineering', name='Shanghai Maritime University', city='Shanghai', country='China'),
        Affiliation(type=None, ref_nr='Aff2', ror='https://ror.org/05k2j8e48', grid='grid.495244.a', isni='0000 0004 1761 5722', division='College of Artificial Intelligence', name='Jiangxi University of Technology', city='Jiangxi', country='China'),
        Affiliation(type=None, ref_nr='Aff3', ror='https://ror.org/04z7qrj66', grid='grid.412518.b', isni='0000 0001 0008 0619', division='College of Merchant Marine', name='Shanghai Maritime University', city='Shanghai', country='China')
        ]
    for a in article:
        assert a.affiliations == expected


def test_article_contributors():
    """Test the contributors of the article."""
    expected = [
        Contributor(type='author', nr='Au1', orcid='http://orcid.org/0000-0002-1606-3511', surname='Hu', given_name='Zhanhui', email=None, affiliations_ref_nr=['Aff1']),
        Contributor(type='author', nr='Au2', orcid=None, surname='Liu', given_name='Guangzhong', email='gzhliu@shmtu.edu.cn', affiliations_ref_nr=['Aff1']),
        Contributor(type='author', nr='Au3', orcid=None, surname='Li', given_name='Yanping', email=None, affiliations_ref_nr=['Aff2']),
        Contributor(type='author', nr='Au4', orcid=None, surname='Zhuang', given_name='Siqing', email=None, affiliations_ref_nr=['Aff3'])
        ]
    for a in article:
        assert a.contributors == expected


def test_article_dates():
    """Test the dates of the article."""
    expected_date_accepted = Date(year=2024, month=7, day=21)
    expected_date_epub = Date(year=2024, month=9, day=9)
    expected_date_online = Date(year=2024, month=9, day=9)
    expected_date_ppub = Date(year=2024, month=12, day=None)
    expected_date_received = Date(year=2024, month=4, day=12)
    expected_date_registration = Date(year=2024, month=7, day=24)

    for a in article:
        assert a.date_accepted == expected_date_accepted
        assert a.date_epub == expected_date_epub
        assert a.date_online == expected_date_online
        assert a.date_ppub == expected_date_ppub
        assert a.date_received == expected_date_received
        assert a.date_registration == expected_date_registration


def test_article_meta():
    """Test the article meta-data."""
    for a in article:
        assert a.article_type == 'research-article'
        assert a.language == 'en'
        assert a.publisher_id == 's40747-024-01577-y'
        assert a.manuscript == '1577'
        assert a.doi == '10.1007/s40747-024-01577-y'
        assert a.title == 'SAGB: self-attention with gate and BiGRU network for intrusion detection'


def test_book_meta():
    """Test the book meta data."""
    for chapter in book:
        assert chapter.book_doi == "10.1007/978-3-031-63501-4"
        assert chapter.publisher_id == "978-3-031-63501-4"
        assert chapter.book_title == "Automated Reasoning"
        assert chapter.book_title_id == "631234"
        assert chapter.book_sub_title == "12th International Joint Conference, IJCAR 2024, Nancy, France, July 3–6, 2024, Proceedings, Part II"
        assert chapter.book_pub_date == "2024-01-01"
        assert chapter.isbn_print == "978-3-031-63500-7"
        assert chapter.isbn_electronic == "978-3-031-63501-4"
        assert chapter.publisher_name == "Springer Nature Switzerland"
        assert chapter.publisher_loc == "Cham"


def test_chapter_affiliations():
    """Test the affiliations of the chapter."""
    expected_affiliations = [
        Affiliation(type='book author', ref_nr='Aff1', ror='https://ror.org/00d7xrm67', grid='grid.410413.3', isni='0000 0001 2294 748X', division='Institute of Software Technology', name='Graz University of Technology', city='Graz', country='Austria'),
        Affiliation(type='book author', ref_nr='Aff2', ror=None, grid='grid.426094.d', isni='0000 0004 0625 6437', division='Corporate Technology', name='Siemens (Austria)', city='Wien', country='Austria'),
        Affiliation(type='book author', ref_nr='Aff3', ror='https://ror.org/03yxnpp24', grid='grid.9224.d', isni='0000 0001 2168 1229', division='ETS de Ingeniería Informática', name='University of Seville', city='Sevilla', country='Spain'),
        Affiliation(type=None, ref_nr='Aff4', ror='https://ror.org/00d7xrm67', grid='grid.410413.3', isni='0000 0001 2294 748X', division='Institute of Software Technology', name='Graz University of Technology', city='Graz', country='Austria'),
        Affiliation(type=None, ref_nr='Aff5', ror=None, grid='grid.426094.d', isni='0000 0004 0625 6437', division='Corporate Technology', name='Siemens (Austria)', city='Wien', country='Austria'),
        Affiliation(type=None, ref_nr='Aff6', ror='https://ror.org/03yxnpp24', grid='grid.9224.d', isni='0000 0001 2168 1229', division='ETS de Ingeniería Informática', name='University of Seville', city='Sevilla', country='Spain')
        ]
    for one_chapter in chapter:
        assert one_chapter.affiliations == expected_affiliations


def test_chapter_contributors():
    """Test the contributors of the chapter."""
    expected_contributors = [
        Contributor(type='author', nr=None, orcid=None, surname='Felfernig', given_name='Alexander', email='alexander.felfernig@ist.tugraz.at', affiliations_ref_nr=['Aff1']),
        Contributor(type='author', nr=None, orcid=None, surname='Falkner', given_name='Andreas', email='andreas.a.falkner@siemens.com', affiliations_ref_nr=['Aff2']),
        Contributor(type='author', nr=None, orcid=None, surname='Benavides', given_name='David', email='benavides@us.es', affiliations_ref_nr=['Aff3']),
        Contributor(type='author', nr=None, orcid=None, surname='Felfernig', given_name='Alexander', email=None, affiliations_ref_nr=['Aff4']),
        Contributor(type='author', nr=None, orcid=None, surname='Falkner', given_name='Andreas', email=None, affiliations_ref_nr=['Aff5']),
        Contributor(type='author', nr=None, orcid=None, surname='Benavides', given_name='David', email=None, affiliations_ref_nr=['Aff6'])
        ]
    for one_chapter in chapter:
        assert one_chapter.contributors == expected_contributors


def test_chapter_dates():
    """Test the dates of the chapter."""
    expected_date_epub = Date(year=2024, month=6, day=30)
    expected_date_online = Date(year=2024, month=6, day=30)
    expected_date_ppub = Date(year=None, month=None, day=None)
    expected_date_registration = Date(year=2024, month=5, day=27)

    for one_chapter in chapter:
        assert one_chapter.date_epub == expected_date_epub
        assert one_chapter.date_online == expected_date_online
        assert one_chapter.date_ppub == expected_date_ppub
        assert one_chapter.date_registration == expected_date_registration


def test_chapter_meta():
    """Test the chapter meta data."""
    for one_chapter in chapter:
        assert one_chapter.doi == '10.1007/978-3-031-61874-1_5'
        assert one_chapter.chapter_nr == 5
        assert one_chapter.title == 'Tools and Applications'

    assert isinstance(book[0], Chapter)


def test_iterable():
    """Test the lengths"""
    assert len(book) == 2
    for book_chapter in book:
        assert isinstance(book_chapter, Chapter)

    assert len(journal) == 3
    for journal_article in journal:
        assert isinstance(journal_article, Article)


def test_journal_meta():
    """Test the journal meta-data."""
    for article in journal:
        assert article.journal_publisher_id == '40747'
        assert article.journal_doi is None
        assert article.journal_title == 'Complex & Intelligent Systems'
        assert article.journal_abbrev_title == 'Complex Intell. Syst.'
        assert article.issn_print == '2199-4536'
        assert article.issn_electronic == '2198-6053'
        assert article.publisher_name == 'Springer International Publishing'
        assert article.publisher_loc == 'Cham'


def test_paragraphs():
    """Test the paragraphs."""
    assert len(chapter[0].paragraphs) == 1
    expected_paragraph = Paragraph(paragraph_id='Par1',
                                            section_id='Abs1',
                                            section_title='Abstract',
                                            text='Feature Models (FMs) are not only an active scientific topic but they are supported by many tools from industry and academia. In this chapter, we provide an overview of example feature modelling tools and corresponding FM configurator applications. In our discussion, we first focus on different tools supporting the design of FMs. Thereafter, we provide an overview of tools that also support FM analysis. Finally, we discuss different existing FM configurator applications.')
    assert chapter[0].paragraphs[0] == expected_paragraph

    assert len(chapter_with_text[0].paragraphs) == 49
    expected_paragraph = Paragraph(paragraph_id='Par5',
                                            section_id='Sec1',
                                            section_title='Introduction',
                                            text='In this work, I aim to provide a better understanding of the governanceGovernance/human factors that Data Scientist and organizations should be aware of. To address this challengeChallenges, I will answer fundamental research questions for the domain.')
    assert chapter_with_text[0].paragraphs[4] == expected_paragraph

    assert len(article[0].paragraphs) == 70
    expected_paragraph = Paragraph(paragraph_id='Par5',
                                            section_id='Sec1',
                                            section_title='Introduction',
                                            text="With the diversification and complexity of intrusion methods and the maturity of machine learning algorithms, studying intrusion detection based on machine learning is a hot topic for scientists. The methods based on machine learning include those based on behavior and statistics, whose main idea is to extract the features of the network's traffic and then use traffic features for modeling and training to obtain a classification model that can identify and classify the new traffic in real-time, which has relatively low complexity and good performance, but still has some shortcomings, mainly reflected in:The normal traffic data is more than abnormal traffic for the training sample, and the data imbalance problem will lead to low precision.Most studies emphasize the model's precision in the data set while neglecting to pay attention to the detection rate of the minority attack categories. However, the attacks of the minority categories often cause more damage to the network than the majority categories' attacks.Deep learning [4] is an effective method to solve the shortcomings of shallow machine learning networks. The CNN and RNN have been widely used in classification problems [5] and regression problems [6]. Some methods based on RNN can learn temporal sequence features, but serial-based sequence training has low convergence efficiency and low accuracy.Some machine learning, such as Random Forest (RF) and Decision tree (DT), only learn shallow features, whose effect of multiple classification precision is not good. Deep learning, such as CNN and RNN, made some achievements, but for some large datasets, the generalization ability is poor and can not be calculated in parallel.")
    assert article[0].paragraphs[4] == expected_paragraph


def test_pagination():
    """Test the pagination."""
    assert len(book_pagination) == 27
    dois = set([chapter.doi for chapter in book_pagination])
    assert len(dois) == 27

    assert len(journal_pagination) == 26
    dois = set([article.doi for article in journal_pagination])
    assert len(dois) == 26



'''Constants for the sprynger package.'''
from pathlib import Path

BASE_URL = 'http://api.springernature.com'

CONFIG_FILE = Path.home()/'.config'/'sprynger'/'sprynger.cfg'

BASE_PATH = Path.home()/'.cache'/'sprynger'
DEFAULT_PATHS = {
    'Metadata': BASE_PATH/'metadata',
    'Meta': BASE_PATH/'meta',
    'OpenAccess': BASE_PATH/'open_access'
}

ONLINE_API = {
    'Metadata': 'metadata',
    'Meta': '/meta/v2',
    'OpenAccess': 'openaccess'
}

FORMAT = {
    'Metadata': 'json',
    'Meta': 'json',
    'OpenAccess': 'jats'
}

LIMIT = {
    'Basic': {
        'Metadata': 25,
        'Meta': 25,
        'OpenAccess': 20,
        'OpenAccessJournal': 20,
        'OpenAccessBook': 20,
    },
    'Premium': {
        'Metadata': 25,
        'Meta': 25,
        'OpenAccess': 20,
        'OpenAccessJournal': 20,
        'OpenAccessBook': 20,
    },
}

VALID_FIELDS = {
    "doi": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "subject": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "keyword": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "language": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "pub": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "year": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "onlinedate": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "onlinedatefrom": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "onlinedateto": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "datefrom": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "dateto": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "dateloaded": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "dateloadedfrom": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "dateloadedto": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "country": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "isbn": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "issn": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "journalid": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "topicalcollection": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    },
    "journalonlinefirst:true": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "date": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "issuetype": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    },
    "issue": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "volume": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "type": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Basic", "Premium"]
    },
    "ContainsElements": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "excludeElements": {
        "api": ["OpenAccess"],
        "plan": ["Premium"]
    },
    "Exclude:Bibliography": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "grid": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "orcid": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "bookdoi": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "latest issue": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    },
    "earliest issue": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    },
    "openaccess:true": {
        "api": ["Meta"],
        "plan": ["Premium"]
    },
    "free:true": {
        "api": ["Meta"],
        "plan": ["Premium"]
    },
    "title:": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "orgname:": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "journal:": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "book:": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "name:": {
        "api": ["Metadata", "OpenAccess", "Meta"],
        "plan": ["Premium"]
    },
    "sort:date": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    },
    "sort:sequence": {
        "api": ["Metadata", "Meta"],
        "plan": ["Premium"]
    }
}



"""Utility functions for parsing data."""
from functools import reduce
from typing import Optional

from lxml.etree import _Element


def get_attr(node: Optional[_Element],
             tag: str,
             attr: str,
             value: str) -> Optional[str]:
    """Get the attribute of a tag in an XML node."""
    if node is not None:
        if node.find(f'.//{tag}[@{attr}="{value}"]') is not None:
            return node.find(f'.//{tag}[@{attr}="{value}"]').text
    return None


def get_text(node: Optional[_Element],
             path: str) -> Optional[str]:
    """Get the text of an XML node."""
    if node is not None:
        if node.find(path) is not None:
            return node.find(path).text
    return None


def make_int_if_possible(val):
    """Attempt a conversion to int type."""
    try:
        return int(val)
    except (TypeError, ValueError):
        return val


def str_to_bool(val):
    """Convert a string to a boolean."""
    if val.lower() in ['true', 't', 'yes', 'y', '1']:
        return True
    elif val.lower() in ['false', 'f', 'no', 'n', '0']:
        return False
    else:
        return val


"""Test the startup of the sprynger package."""
from pathlib import Path

from sprynger import init
from sprynger.utils.startup import CUSTOM_KEYS, get_config, get_keys


def test_empty_keys():
    """Test the empty keys."""
    assert CUSTOM_KEYS is None


def test_init():
    """Test the init function."""
    init()
    assert get_config() is not None
    assert get_keys() is not None


def test_create_config():
    """Test the create_config function."""
    config_dir = Path.home()/'.config'/'sprynger'/'test_sprynger.cfg'
    # Dete file
    config_dir.unlink(missing_ok=True)
    assert not config_dir.exists()
    # Create file
    init(config_dir, ['test_key'])
    assert config_dir.exists()



"""Module to test exceptions and warnings."""
import pytest

from sprynger import Metadata, init
from sprynger.exceptions import AuthenticationError

def test_empty_response():
    """Test the empty response."""
    init()
    with pytest.warns(UserWarning, match='No results where found. Check the query.'):
        Metadata(doi='does_not_exist', refresh=True)

def test_authentication_error():
    """Test the authentication error."""
    init(keys=['does_not_exist'])
    with pytest.raises(AuthenticationError, match='Authentication failed. Check your API key.'):
        Metadata('10.1007/s10660-023-09761-x', refresh=True)



"""Import all the necessary classes and functions for the user to use the package."""
from sprynger.meta import Meta
from sprynger.metadata import Metadata
from sprynger.openaccess import OpenAccess
from sprynger.utils.startup import init



"""
Module with Meta class.
"""
from typing import Union

from sprynger.metadata import Metadata
from sprynger.utils.data_structures import MetadataCreator, MetaDiscipline, MetaRecord, MetaURL
from sprynger.utils.parse import make_int_if_possible, str_to_bool


class Meta(Metadata):
    """Class to retreive the metadata of a document from the Springer Meta v2 API."""
    @property
    def records(self) -> list[MetaRecord]:
        """Contains the individual records that matched the query.

        Returns:
            list[MetaRecord]: List of MetaRecord objects which contain the following
            items of a document: `contentType`,
            `identifier`, `language`, `urls`, `title`, `creators`,
            `publicationName`, `openaccess`, `doi`, `publisher`, `publicationDate`,
            `publicationType`, `issn`, `eIssn`, `volume`, `number`, `issueType`,
            `topicalCollection`, `genre`, `startingPage`,
            `endingPage`, `journalId`, `onlineDate`,
            `copyright`, `abstract`, `conferenceInfo`, 
            `keyword`, `subjects` and `disciplines`.
        """
        records_list = []
        for record in self._json.get('records', []):
            # Parse the URLs
            url_list = []
            for url in record.get('url', []):
                url_format = url.get('format')
                platform = url.get('platform')
                value = url.get('value')
                url_list.append(MetaURL(format=url_format, platform=platform, value=value))
            # Parse the creators
            creators = []
            for ceator in record.get('creators', []):
                creators.append(MetadataCreator(creator=ceator.get('creator'),
                                            ORCID=ceator.get('ORCID')))
            # Parse the disciplines
            disciplines = []
            for discipline in record.get('disciplines', []):
                disciplines.append(
                    MetaDiscipline(id=discipline.get('id'),
                                   term=discipline.get('term'))
                                   )

            records_list.append(
                MetaRecord(
                    contentType=record.get('contentType'),
                    identifier=record.get('identifier'),
                    language=record.get('language'),
                    urls=url_list,
                    title=record.get('title'),
                    creators=creators,
                    publicationName=record.get('publicationName'),
                    openaccess=str_to_bool(record.get('openaccess')),
                    doi=record.get('doi'),
                    publisher=record.get('publisher'),
                    publicationDate=record.get('publicationDate'),
                    publicationType=record.get('publicationType'),
                    issn=record.get('issn'),
                    eIssn=record.get('eIssn'),
                    volume=make_int_if_possible(record.get('volume')),
                    number=make_int_if_possible(record.get('number')),
                    issueType=record.get('issueType'),
                    topicalCollection=record.get('topicalCollection'),
                    genre=record.get('genre'),
                    startingPage=make_int_if_possible(record.get('startingPage')),
                    endingPage=make_int_if_possible(record.get('endingPage')),
                    journalId=make_int_if_possible(record.get('journalId')),
                    onlineDate=record.get('onlineDate'),
                    copyright=record.get('copyright'),
                    abstract=record.get('abstract'),
                    conferenceInfo = record.get('conferenceInfo'),
                    keyword = record.get('keyword'),
                    subjects=record.get('subjects'),
                    disciplines=disciplines
                )
            )
        return records_list

    def __init__(self,
                 query: str = '',
                 start: int = 1,
                 nr_results: int = 10,
                 premium: bool = False,
                 cache: bool = True,
                 refresh: Union[bool, int] = False,
                 **kwargs):
        """
        Args:
            query (str): The query to search for.
            start (int): The starting index for the results. Defaults to 1.
            nr_results (int): The number of results to retrieve. Defaults to 10.
            premium (bool): Whether the user has a premium account. Defaults to False.
            cache (bool): Whether to cache the results. Defaults to True.
            refresh (bool|int): Weather to refresh the cache. If an integer is provided, 
                it will be used as the cache expiration time in days. Defaults to False.
            kwargs: Additional fields for query (e.g. issn, datefrom, dateto, etc.). For a comprehensive list of
                available fields, see the 
                `Springer Metadata API documentation <https://docs-dev.springernature.com/docs/#querying-api/querying-api>`_.

        This class is iterable, allowing you to iterate over the metadata `records` retrieved.
        It also supports indexing to access the metadata of specific documents.

        Example:
            Retrieve the metadata of documents that match the query 'Segmentation' with ISSN '1573-7497' and
            starting from the date '2024-01-01'.
            
            >>> meta = Meta('Segmentation', issn='1573-7497', datefrom='2024-01-01')
            >>> for record in metadata:
            >>>     print(record)
     
        Note:
            The properties `facets`, `records` and `results` can be converted to a pandas
            DataFrame with `pd.DataFrame(object.property)`.
        """
        super().__init__(query=query,
                         start=start,
                         nr_results=nr_results,
                         premium=premium,
                         cache=cache,
                         refresh=refresh,
                         **kwargs)
        self._nr_results = nr_results
        self._records = self.records



"""Base class to retrieve data from the Springer API."""
from __future__ import annotations
from math import ceil
import os
import json
from json.decoder import JSONDecodeError
from typing import Optional, Literal, Union
from datetime import datetime, timedelta
import warnings

from lxml import etree
from requests import Response

from sprynger.utils.constants import BASE_URL, FORMAT, LIMIT, ONLINE_API
from sprynger.utils.fetch import fetch_data
from sprynger.utils.startup import get_config, get_keys

class Base:
    """Base class to retrieve data from the Springer API."""
    @property
    def _json(self) -> dict:
        """JSON response from the API."""
        return _to_json(self._res)

    @property
    def _xml(self) -> etree._Element:
        """XML response from the API."""
        return _to_xml(self._res)

    def __init__(self,
                 query: str,
                 api: Literal['Metadata', 'Meta', 'OpenAccess'],
                 start: int = 1,
                 nr_results: int = 10,
                 premium: bool = False,
                 cache: bool = True,
                 refresh: Union[bool, int] = False) -> None:

        config = get_config()

        self._api = api
        online_api = ONLINE_API[api]
        rate_limit = LIMIT['Premium'][api] if premium else LIMIT['Basic'][api]
        limit = min(nr_results, rate_limit)

        keys = get_keys()
        self._url = f'{BASE_URL}/{online_api}/{FORMAT[api]}'
        self._params = {'q': query,
                        's': start,
                        'p': limit,
                        'api_key': keys[0]}

        # Generate a cache key based on the parameters
        cache_dir = config.get('Directories', api)
        self._cache_key = self._create_cache_key(query, start, limit)
        self._cache_file = os.path.join(cache_dir, f'{self._cache_key}.{FORMAT[api]}')
        self._refresh = refresh
        self._cache = cache

        self._res = self._fetch_or_load()
        n_found = self._get_total_results()

        if n_found == 0:
            warnings.warn('No results where found. Check the query.', UserWarning)

        n = min(nr_results, n_found)
        n_chunks = ceil(n / limit)
        for i in range(1, n_chunks):
            new_start = start + i * limit
            n -= limit
            limit = min(n, limit)
            # Update variables for querying the next chunk
            self._params.update({'s': new_start,
                                 'p': limit})
            self._cache_key = self._create_cache_key(query, new_start, limit)
            self._cache_file = os.path.join(cache_dir, f'{self._cache_key}.{FORMAT[api]}')
            tmp_res = self._fetch_or_load()
            self._res = self._append_response(tmp_res)


    def _create_cache_key(self, query: str, start: int, limit: int) -> str:
        """Create a cache key based on the query and start."""
        cache_key = f'{query}_{start}_{limit}'
        for old, new in [(':', '-'), ('/', '-')]:
            cache_key = cache_key.replace(old, new)
        return cache_key


    def _append_response(self, tmp_res: Union[Response, MockResponse]):
        """Append the response to the current response."""
        if FORMAT[self._api] == 'json':
            data_json = _to_json(self._res)
            tmp_res_json = _to_json(tmp_res)
            # Append the records from the tmp response to the current response
            data_json['records'].extend(tmp_res_json.get('records', []))
            return MockResponse(data_json)
        elif FORMAT[self._api] == 'jats':
            data_xml = _to_xml(self._res)
            tmp_res_xml = _to_xml(tmp_res)
            # Append the records from the tmp response to the current response
            data_records = data_xml.find('./records')
            tmp_res_records = tmp_res_xml.find('./records')
            for record in tmp_res_records:
                data_records.append(record)
            data_xml_str = etree.tostring(data_xml, encoding='unicode')
            return MockResponse(data_xml_str, is_xml=True)
        else:
            raise ValueError(f'Unknown format: {FORMAT[self._api]}')

    def _fetch_or_load(self) -> Union[Response, MockResponse]:
        """Fetch or load the data from the cache."""
        if self._should_fetch():
            res = self._fetch()
        else:
            res = self._load_from_cache()
        return res

    def _get_total_results(self):
        """Get the total number of results for the query."""
        if (self._api == 'Metadata') or (self._api == 'Meta'):
            res_json = _to_json(self._res)
            total = res_json['result'][0]['total']
        elif self._api in ['OpenAccess']:
            res_xml = _to_xml(self._res)
            total = res_xml.find('./result/total').text
        else:
            raise ValueError(f'Unknown API: {self._api}')
        return int(total)

    def _should_fetch(self) -> bool:
        """Determine whether the data has to be fetched."""
        if isinstance(self._refresh, bool) and self._is_cached():
            return self._refresh # If is cached user decides to fetch
        elif isinstance(self._refresh, int) and self._is_cached():
            cache_age = datetime.now() - datetime.fromtimestamp(os.path.getmtime(self._cache_file))
            return cache_age > timedelta(days=self._refresh)  #Fetch if cache is older than specified days
        return True  # If no cache exists, return True to fetch

    def _is_cached(self) -> bool:
        """Check if the cache file exists."""
        return os.path.exists(self._cache_file)

    def _load_from_cache(self) -> MockResponse:
        """Load response from the cache."""
        with open(self._cache_file, 'r') as f:
            if FORMAT[self._api] == 'json':
                return MockResponse(json.load(f))
            elif FORMAT[self._api] == 'jats':  # XML-based format
                return MockResponse(f.read(), is_xml=True)
            else:
                raise ValueError(f'Unknown format: {FORMAT[self._api]}')

    def _fetch(self) -> Response:
        """Fetch data from the API and cache the response."""
        res = fetch_data(url=self._url, params=self._params)
        if self._cache:
            # Save the response to the cache file depending on format
            with open(self._cache_file, 'w') as f:
                if FORMAT[self._api] == 'json':
                    json.dump(res.json(), f)
                elif FORMAT[self._api] == 'jats':  # XML-based format
                    f.write(res.content.decode())
        return res

def _to_json(response) -> dict:
    """Auxiliary method to convert the response to JSON."""
    try:
        return response.json()
    except JSONDecodeError:
        return {}

def _to_xml(response) -> etree._Element:
    """Auxiliary method to convert the response to XML."""
    return etree.fromstring(text=response.content)


class MockResponse:
    """Mock response class for cached data."""
    def __init__(self, data: Union[dict, str], is_xml: bool = False) -> None:
        """Initialize the cached response, either JSON or XML (JATS)."""
        self._data = data
        self.is_xml = is_xml

    def json(self) -> Optional[dict]:
        """Return the JSON data if it's JSON."""
        if not self.is_xml:
            return self._data
        return None

    @property
    def content(self) -> str:
        """Return the raw content (used for XML parsing if needed)."""
        if self.is_xml:
            return self._data
        return json.dumps(self._data)



"""Utility functions for fetching data from the Springer API."""
from requests.adapters import HTTPAdapter
from requests import Response, Session
from urllib3.util.retry import Retry

from sprynger.utils.startup import get_config

from sprynger.exceptions import (
    APIError,
    AuthenticationError,
    InvalidRequestError,
    ResourceNotFoundError,
)


def create_session(max_retries: int,
                   backoff_factor: float) -> Session:
    """Create a session."""
    session = Session()
    retries = Retry(total=max_retries,
        backoff_factor=backoff_factor,
        backoff_max=60,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["GET"]
    )
    adapter = HTTPAdapter(max_retries=retries)
    session.mount("https://", adapter)
    session.mount("http://", adapter)
    return session


def check_response(response: Response) -> None:
    """Check the response."""
    status_code = response.status_code
    if status_code != 200:
        if response.status_code == 400:
            raise InvalidRequestError(response.status_code)
        elif response.status_code == 401 or response.status_code == 403:
            raise AuthenticationError(response.status_code)
        elif response.status_code == 404:
            raise ResourceNotFoundError(response.status_code)
        else:
            raise APIError(response.status_code, "Unhandled error occurred")


def fetch_data(url: str, params: dict) -> Response:
    """Fetch data from the Springer API."""
    # Get the configuration
    config = get_config()
    max_retries = config.getint('Requests', 'Retries', fallback=5)
    backoff_factor = config.getfloat('Requests', 'BackoffFactor', fallback=2.0)
    timeout = config.getint('Requests', 'Timeout', fallback=20)

    # Create session and retrieve data
    session = create_session(max_retries, backoff_factor)
    response = session.get(url, params=params, timeout=timeout)
    check_response(response)

    return response



"""
Module with the OpenAccess class to retrieve Open Access documents from Springer Nature API.
Iterate or slice the OpenAccess object to get the documents. Documents can be either:

- Articles: OpenAccess articles from journals.
- Chapters: OpenAccess chapters from books.

Example:
    >>> oa = OpenAccess('Eigenvalues', datefrom='2024-01-01')
    >>> for doc in oa:
    >>>     print(doc)
"""
from typing import Optional, Union

from sprynger.retrieve import Retrieve
from sprynger.openaccess_article import Article
from sprynger.openaccess_chapter import Chapter


class OpenAccess(Retrieve):
    """Retrieve Open Access documents from Springer Nature API."""
    @property
    def documents_found(self) -> int:
        """Number of documents found."""
        return self._get_total_results()

    def _get_documents(self) -> list[Union[Chapter, Article]]:
        """Auciliary method to retrieve the documents from the Open Access API."""
        documents = []
        for record in self._xml.find('.//records'):
            if record.tag == 'book-part-wrapper':
                documents.append(Chapter(record))
            elif record.tag == 'article':
                documents.append(Article(record))
            else:
                raise ValueError(f'Unknown document type: {record.tag}')
        return documents

    @property
    def xml(self) -> str:
        """Raw XML response from the Open Access API."""
        return self._xml

    def __init__(
        self,
        query: str = '',
        start: int = 1,
        nr_results: int = 10,
        premium: bool = False,
        cache: bool = True,
        refresh: Union[bool, int] = False,
        **kwargs,
    ) -> None:
        """
        Args:
            query (str): Query string.
            start (int): Start index of the results.
            nr_results (int): Number of results to retrieve.
            premium (bool): Use the premium API.
            cache (bool): Use the cache.
            refresh (Union[bool, int]): Refresh the cache.
            **kwargs: Additional fields for query (e.g. issn, datefrom, dateto, etc.).
                For a comprehensive list of available fields, see the 
                `Springer Metadata API documentation <https://docs-dev.springernature.com/docs/#querying-api/querying-api>`_.
        
        Example:
            Retrieve Open Access articles from the journal with ISSN '2223-7704' published after 
            '2024-01-01'.

            >>> oa = OpenAccess(issn=2223-7704, datefrom='2024-01-01')
        """

        super().__init__(query=query,
                         api='OpenAccess',
                         start=start,
                         nr_results=nr_results,
                         premium=premium,
                         cache=cache,
                         refresh=refresh,
                         **kwargs)
        self.documents = self._get_documents()

    def __iter__(self):
        return iter(self.documents)

    def __getitem__(self, index):
        return self.documents[index]

    def __len__(self):
        return len(self.documents)

    def __repr__(self):
        return self.documents.__repr__()



"""Module with the Exceptions Classes"""

class APIError(Exception):
    """Base class for API errors"""
    def __init__(self, status_code, message="An error occurred with the API request"):
        self.status_code = status_code
        self.message = message
        super().__init__(f"{status_code}: {message}")


class AuthenticationError(APIError):
    """Exception raised for 401/403 Authentication Failures"""
    def __init__(self, status_code, message="Authentication failed. Check your API key."):
        super().__init__(status_code, message)


class RateLimitError(APIError):
    """Exception raised for 429 Rate Limit Exceeded"""
    def __init__(self, status_code, message="Rate limit exceeded. Try again later."):
        super().__init__(status_code, message)


class ResourceNotFoundError(APIError):
    """Exception raised for 404 Resource Not Found"""
    def __init__(self, status_code, message="Resource not found. Check your URL or query."):
        super().__init__(status_code, message)


class InvalidRequestError(APIError):
    """Exception raised for 400 Bad Request due to invalid formats"""
    def __init__(self, status_code, message="Invalid request format. Check your parameters."):
        super().__init__(status_code, message)


class InternalServerError(APIError):
    """Exception raised for 500 Internal Server Error"""
    def __init__(self, status_code, message="Internal server error. Try again later."):
        super().__init__(status_code, message)



"""Module with the Article class for the OpenAccess class."""
from typing import Optional

from sprynger.utils.data_structures import Affiliation, Contributor, Date, Paragraph
from sprynger.utils.parse import get_attr, get_text
from sprynger.utils.parse_openaccess import (
    affs_to_dict,
    get_contributors,
    get_affiliations,
    get_date,
    get_paragraphs,
)

class Article:
    """Auxiliary class to parse an article from a journal."""
    @property
    def affiliations(self) -> list[Affiliation]:
        """List of affiliations of the collaborators of the article. Each affiliation is represented
        as a named tuple with the following fields:
        `type`, `ref_nr`, `ror`, `grid`, `isni`, `division`, `name`, `city`, `country`.
        
        Note: To match affiliations with contributors use the affiliation's `ref_nr` and the
        contributor's `affiliations_ref_nr`.
        """
        return get_affiliations(self._data)

    @property
    def affiliations_dict(self) -> dict[str, Affiliation]:
        """Auxiliary property to query the affiliations by their reference number."""
        return affs_to_dict(self.affiliations)

    @property
    def article_type(self) -> Optional[str]:
        """Type of the article."""
        return self._data.get('article-type')

    @property
    def contributors(self) -> list[Contributor]:
        """List of contributors of the article. Each contributor is represented as a named tuple
        with the following fields:
        `type`, `nr`, `orcid`, `surname`, `given_name`, `email`, `affiliations_ref_nr`.

        Note: To match contributors with affiliations use the contributor's `affiliations_ref_nr`
        and the affiliation's `ref_nr`.
        """
        return get_contributors(self._article_meta)

    @property
    def date_epub(self) -> Date:
        """Electronic publication date of the article."""
        date_node = self._article_meta.find('.//pub-date[@publication-format="electronic"]')
        return get_date(date_node)

    @property
    def date_ppub(self) -> Date:
        """Print publication date of the article."""
        date_node = self._article_meta.find('.//pub-date[@publication-format="print"]')
        return get_date(date_node)

    @property
    def date_registration(self) -> Date:
        """Registration date of the article."""
        date_node = self._article_meta.find('.//history/date[@date-type="registration"]')
        return get_date(date_node)

    @property
    def date_received(self) -> Date:
        """Date when article was recieved."""
        date_node = self._article_meta.find('.//history/date[@date-type="received"]')
        return get_date(date_node)

    @property
    def date_accepted(self) -> Date:
        """Accepted date of the article."""
        date_node = self._article_meta.find('.//history/date[@date-type="accepted"]')
        return get_date(date_node)

    @property
    def date_online(self) -> Date:
        """Online date of the article."""
        date_node = self._article_meta.find('.//history/date[@date-type="online"]')
        return get_date(date_node)

    @property
    def doi(self) -> Optional[str]:
        """DOI of the article."""
        return get_attr(self._article_meta, 'article-id', 'pub-id-type', 'doi')

    @property
    def issn_electronic(self) -> Optional[str]:
        """Electronic ISSN of the journal."""
        return get_attr(self._journal_meta, 'issn', 'pub-type', 'epub')

    @property
    def issn_print(self) -> Optional[str]:
        """Print ISSN of the journal."""
        return get_attr(self._journal_meta, 'issn', 'pub-type', 'ppub')

    @property
    def journal_abbrev_title(self) -> Optional[str]:
        """Abbreviated title of the journal."""
        return get_text(self._journal_meta, './/abbrev-journal-title')

    @property
    def journal_doi(self) -> Optional[str]:
        """DOI of the journal."""
        return get_attr(self._journal_meta, 'journal-id', 'journal-id-type', 'doi')

    @property
    def journal_publisher_id(self) -> Optional[str]:
        """Publisher ID of the journal."""
        return get_attr(self._journal_meta, 'journal-id', 'journal-id-type', 'publisher-id')

    @property
    def journal_title(self) -> Optional[str]:
        """Title of the journal."""
        return get_text(self._journal_meta, './/journal-title')

    @property
    def language(self) -> Optional[str]:
        """Language of the article."""
        return self._data.get('{http://www.w3.org/XML/1998/namespace}lang')

    @property
    def manuscript(self) -> Optional[str]:
        """Manuscript of the article."""
        return get_attr(self._article_meta, 'article-id', 'pub-id-type', 'manuscript')

    @property
    def paragraphs(self) -> list[Paragraph]:
        """Paragraphs of the article.

        Returns:
            list[Paragraph]: A list of Paragraph objects containing the
            `paragraph_id`, `section_id`, `section_title`, and `text`.
        """
        return get_paragraphs(self._data)

    @property
    def publisher_id(self) -> Optional[str]:
        """Publisher ID of the article."""
        return get_attr(self._article_meta, 'article-id', 'pub-id-type', 'publisher-id')

    @property
    def publisher_loc(self) -> Optional[str]:
        """Location of the publisher."""
        return get_text(self._journal_meta, './/publisher-loc')

    @property
    def publisher_name(self) -> Optional[str]:
        """Name of the publisher."""
        return get_text(self._journal_meta, './/publisher-name')

    @property
    def title(self) -> Optional[str]:
        """Title of the article."""
        return get_text(self._article_meta, './/title-group/article-title')

    def __init__(self, data):
        self._data = data
        self._journal_meta = data.find('.//front/journal-meta')
        self._article_meta = data.find('.//front/article-meta')

    def __repr__(self) -> str:
        return f'Article {self.doi}'



"""Tests for the Meta class."""
from sprynger import init
from sprynger import Metadata
from sprynger.utils.data_structures import (MetadataCreator,
                                            MetadataFacets,
                                            MetadataRecord,
                                            MetadataResult)

init()

journal_metadata = Metadata(issn='3004-9261', start=1, nr_results=2, refresh=30)
book_metadata = Metadata(isbn='978-1-0716-1418-1', start=1, nr_results=3, refresh=30)
meta_pagination = Metadata('issn:2662-9984', nr_results=30, refresh=True)
article_metadata = Metadata('doi:10.1186/s43593-023-00053-3', refresh=30)


def test_book():
    """Test the book metadata."""
    for book in book_metadata.records:
        assert book.contentType == 'Chapter'
        assert book.publicationName == 'An Introduction to Statistical Learning'


def test_journal():
    """Test the journal metadata."""
    for article in journal_metadata:
        assert article.publicationName == 'Discover Applied Sciences'
        assert article.issn == '3004-9261'
        assert article.journalId == 42452


def test_len():
    """Test the length of the metadata."""
    assert len(article_metadata.records) == 1

    assert len(journal_metadata.records) == 2
    assert len(journal_metadata) == 2

    assert len(book_metadata.records) == 3
    assert len(book_metadata) == 3


def test_factets():
    """Test the facets."""
    assert len(article_metadata.facets) == 11

    expected = MetadataFacets(facet='subject', value='Physics', count=1)
    assert article_metadata.facets[1] == expected


def test_records():
    """Test the records."""
    expected = MetadataRecord(contentType='Article',
                              identifier='doi:10.1186/s43593-023-00053-3',
                              language='en',
                              url='http://dx.doi.org/10.1186/s43593-023-00053-3',
                              url_format='',
                              url_platform='',
                              title='Highly efficient flexible structured metasurface by roll-to-roll printing for diurnal radiative cooling',
                              creators=[MetadataCreator(creator='Lin, Keng-Te', ORCID=None),
                                        MetadataCreator(creator='Nian, Xianbo', ORCID=None),
                                        MetadataCreator(creator='Li, Ke', ORCID=None),
                                        MetadataCreator(creator='Han, Jihong', ORCID=None),
                                        MetadataCreator(creator='Zheng, Nan', ORCID=None),
                                        MetadataCreator(creator='Lu, Xiaokang', ORCID=None),
                                        MetadataCreator(creator='Guo, Chunsheng', ORCID=None),
                                        MetadataCreator(creator='Lin, Han', ORCID=None),
                                        MetadataCreator(creator='Jia, Baohua', ORCID='0000-0002-6703-477X')],
                              publicationName='eLight',
                              openaccess=True,
                              doi='10.1186/s43593-023-00053-3',
                              publisher='Springer',
                              publicationDate='2023-10-25',
                              publicationType='Journal',
                              issn='2662-8643',
                              volume=3,
                              number=1,
                              genre=['OriginalPaper', 'Research Article'],
                              startingPage=1,endingPage=12, journalId=43593,
                              copyright='©2023 The Author(s)',
                              abstract='An ideal radiative cooler requires accurate spectral control capability to achieve efficient thermal emission in the atmospheric transparency window (8–13\xa0μm), low solar absorption, good stability, scalability, and a simple structure for effective diurnal radiative cooling. Flexible cooling films made from polymer relying on polymer intrinsic absorbance represent a cost-effective solution but lack accuracy in spectral control. Here, we propose and demonstrate a metasurface concept enabled by periodically arranged three-dimensional (3D) trench-like structures in a thin layer of polymer for high-performance radiative cooling. The structured polymer metasurface radiative cooler is manufactured by a roll-to-roll printing method. It exhibits superior spectral breadth and selectivity, which offers outstanding omnidirectional absorption/emission (96.1%) in the atmospheric transparency window, low solar absorption (4.8%), and high stability. Impressive cooling power of 129.8 W m −2 and temperature deduction of 7\xa0°C on a clear sky midday have been achieved, promising broad practical applications in energy saving and passive heat dispersion fields.',
                              subjects=['Physics', 'Optics, Lasers, Photonics, Optical Devices'])
    assert article_metadata.records[0] == expected


def test_pagination():
    """Test the pagination."""
    assert len(meta_pagination.records) == 30
    ids = set([record.identifier for record in meta_pagination.records])
    assert len(ids) == 30


def test_results():
    """Test the results."""
    expected = MetadataResult(total=1, start=1, pageLength=10, recordsRetrieved=1)
    assert article_metadata.results == expected
